//------------------------------------------------------------------------
//Model.cpp	
//------------------------------------------------------------------------
#include "stdafx.h"
#include "VixarLibrary.h"
#include <map>

//------------------------------------------------------------------------
//CModelNX
//------------------------------------------------------------------------
CModel::CModel()
{
	m_vertices.clear();
	m_faces.clear();
	m_animations.clear();
	m_mirror_axis = 0;
	m_facet = 0;
	m_vb = NULL;
	D3DXMatrixIdentity(&m_transform);
}

//------------------------------------------------------------------------
//fXgN^
//------------------------------------------------------------------------
CModel::~CModel()
{
}

//------------------------------------------------------------------------
//Zq=̃I[o[[h
//------------------------------------------------------------------------
CModel& CModel::operator=(const CModel& pt)
{
	m_name = pt.m_name;
	m_vertices = pt.m_vertices;
	m_faces = pt.m_faces;
	m_groups = pt.m_groups;
	m_animations = pt.m_animations;
	m_mirror_axis = pt.m_mirror_axis;
	m_vb = pt.m_vb;
	m_transform = pt.m_transform;
	m_facet = pt.m_facet;

	return *this;
}

//------------------------------------------------------------------------
//Aj[V̎擾
//------------------------------------------------------------------------
D3DXMATRIX CModel::GetAnimation(int i,int j)
{
	//w肵Aj[V̍s擾
	int size = (int)m_animations.size();
	if ( size > 0 && i >= 0 && i < size )
	{
		int size2 = (int)m_animations[i].size();
		if ( size2 > 0 && j >= 0 && j < size2 )
		{
			return m_animations[i][j];
		}
	}

	//Aj[VȂꍇPʍsԂ
	return D3DXMATRIX(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1);
}

//------------------------------------------------------------------------
//animations̃TCY̎擾
//------------------------------------------------------------------------
int CModel::GetAnimationSize(int i)
{
	int size = (int)m_animations.size();
	if ( size >= 0 && i < size )
	{
		return (int)m_animations[i].size();
	}

	return 0;
}

//------------------------------------------------------------------------
//ʂƂ̖@OςvZ
//------------------------------------------------------------------------
void CModel::CalcFlatFaceNormal()
{
	D3DXVECTOR3 a,b;

	for( vector<CFace>::iterator face = m_faces.begin(); face != m_faces.end(); ++face )
	{
		//ʂ̓ӂ̕xNg
		a = m_vertices[face->GetIndex(0)] - m_vertices[face->GetIndex(1)];
		b = m_vertices[face->GetIndex(0)] - m_vertices[face->GetIndex(2)];
		//ӂ̕xNg̊Oς@vZ
		face->CrossProduct(a,b);
		//@̐K
		face->Normalize();
	}
}

//------------------------------------------------------------------------
//_Ƃ̖@vZ	
//------------------------------------------------------------------------
void CModel::CalcFaceVertexNormal()
{
	// ʂƂ̒_Ƃ̖@
	for( vector<CFace>::iterator face = m_faces.begin(); face != m_faces.end(); ++face )
	{
		for ( int i = 0; i < face->GetVertexSize(); ++i )
		{
			face->SetVNormal(i,D3DXVECTOR3(0,0,0));

			for ( int f = 0; f < face->GetHaveFacesSize(i); ++f )
			{
				face->AddVNormal(i,m_faces[face->GetHaveFaces(i,f)].GetNormal());
			}
			face->VNormalize(i);
		}
	}
}

//------------------------------------------------------------------------
//_obt@̏
//------------------------------------------------------------------------
void CModel::CreateVertexBuffer(LPDIRECT3DDEVICE9 dev,std::vector<CMaterial> &materials)
{
	int i, size = 0;
	NVERTEX *nvertex;

	//e_ʂ̎擾
	FacesVertexBelongsTo();
	//_Ƃ̖@vZ
	CalcFaceVertexNormal();

	//_v
	for( vector<CFace>::iterator face = m_faces.begin(); face != m_faces.end(); ++face )
	{
		switch(face->GetIndexSize())
		{
		case 3: size += 3; break;
		case 4: size += 6; break;
		}
	}
	//_obt@쐬
	dev->CreateVertexBuffer(sizeof(NVERTEX)*size,0,FVF_NVERTEX,D3DPOOL_MANAGED,&m_vb,NULL);
	//}eȂSĂ̖ʂ̃O[vɕ
	m_groups.resize((int)materials.size());
	m_vb->Lock(0,0,(void **)&nvertex,0);
	for ( int m = 0; m < (int)materials.size(); ++m )
	{
		m_groups[m] = 0;
		for( vector<CFace>::iterator face = m_faces.begin(); face != m_faces.end(); ++face )
		{
			//ʂ̃}eAԍmȂꍇ
			if ( face->GetMaterial() != m )
			{
				continue;
			}
			//O[vƂ̖ʐZ
			switch(face->GetIndexSize())
			{
			case 3: m_groups[m] += 1; break;
			case 4: m_groups[m] += 2; break;
			}
			//ʂ̒_Ƃɒ_Zbg
			switch ( face->GetIndexSize() )
			{
			case 3:
				for ( i = 0; i < 3; ++i,++nvertex )
				{
				const D3DXVECTOR3& pos = m_vertices[face->GetIndex(i)];
				const D3DXVECTOR3& normal = face->GetVNormal(i);
				SetNVertex(
					nvertex,
					pos,
					normal,
					face->GetU(i),
					face->GetV(i));
				}
				break;
			case 4:
				for ( i = 0; i < 3; ++i,++nvertex )
				{
					const D3DXVECTOR3& pos = m_vertices[face->GetIndex(i)];
					const D3DXVECTOR3& normal = face->GetVNormal(i);
					SetNVertex(
						nvertex,
						pos,
						normal,
						face->GetU(i),
						face->GetV(i));
				}
				for ( i = 2; i < 5; ++i,++nvertex )
				{
					int j = (i==4)?0:i;
					const D3DXVECTOR3& pos = m_vertices[face->GetIndex(j)];
					const D3DXVECTOR3& normal = face->GetVNormal(j);
					SetNVertex(
						nvertex,
						pos,
						normal,
						face->GetU(j),
						face->GetV(j));
				}
				break;
			}
		}
	}

	m_vb->Unlock();
}

//------------------------------------------------------------------------
//_̍WlȂǂݒ肷֐
//------------------------------------------------------------------------
void SetNVertex(NVERTEX *v,const D3DXVECTOR3& pos,const D3DXVECTOR3& normal,float tu,float tv)
{
	//_xNg
	v->x  = pos.x;
	v->y  = pos.y;
	v->z  = pos.z;
	//@xNg
	v->nx = normal.x;
	v->ny = normal.y;
	v->nz = normal.z;
	//UVW
	v->tu = tu;
	v->tv = tv;
}

//------------------------------------------------------------------------
//e_ʂ̐ݒ
//------------------------------------------------------------------------
void CModel::FacesVertexBelongsTo()
{
	int f,i,v,size;
	vector <CVertex> normals;

	size = (int)m_faces.size();

	//_Ƃɂ̒_L
	normals.resize((int)m_vertices.size());
	for ( f = 0; f < size; ++f )
	{
		CFace *face = &m_faces[f];
		for ( v = 0; v < face->GetIndexSize(); ++v )
		{
			normals[face->GetIndex(v)].PushBackHaveFaces(f);
		}
	}

	CalcFlatFaceNormal();

	//ʂ̒_ƂVertexIndex̐
	for ( f = 0; f < size; ++f )
	{
		m_faces[f].ResizeVertex(m_faces[f].GetIndexSize());
	}

	//ʂ̒_Ƃɂ̒_L
	for ( f = 0; f < size; ++f )
	{
		CFace *face = &m_faces[f];
		for ( i = 0; i < face->GetIndexSize(); ++i )
		{
			CVertex* vert = &normals[face->GetIndex(i)];
			for ( v = 0; v < vert->GetHaveFacesSize(); ++v )
			{
				int f2 = vert->GetHaveFaces(v);
				if ( f == f2 )
				{
					face->PushBackHaveFaces(i,f2);
					continue;
				}
				float angle = (float)acos(D3DXVec3Dot(&face->GetNormal(),&m_faces[f2].GetNormal()));
				if ( angle < D3DXToRadian(m_facet) ) 
				{
					face->PushBackHaveFaces(i,f2);
				}
			}
		}
	}
}

//------------------------------------------------------------------------
//CModel::DrawPrimitive	
//------------------------------------------------------------------------
void CModel::DrawPrimitive(LPDIRECT3DDEVICE9 dev,D3DXMATRIX *matWorld,std::vector<CMaterial> &materials)
{
	int pre_material = -1,num = 0;

	//[hs̃Zbg
	dev->SetTransform(D3DTS_WORLD,matWorld);
	//[hs̎擾(Collisionp)
	dev->GetTransform(D3DTS_WORLD,&m_transform);
	//eNX`ȂɃZbg
	dev->SetTexture(0,NULL);
	// gp钸_obt@̐ݒ
	dev->SetStreamSource(0,m_vb,0,sizeof(NVERTEX));
	//O[v
	for( int i = 0; i < (int)m_groups.size(); ++i )
	{
		//O[vɖʂ݂ꍇ
		if ( m_groups[i] > 0 )
		{
			CMaterial* material = &materials[i];
			dev->SetMaterial(material->GetMaterial());
			// eNX`ݒ
			if ( material->GetTextureName().empty() )
			{
				//eNX`Ȃꍇ
				dev->SetTexture(0,NULL);
			}
			else
			{
				//eNX`ꍇ
				dev->SetTexture(0,material->GetTexture());
			}
			//v~eBu`
			dev->DrawPrimitive(D3DPT_TRIANGLELIST,num,m_groups[i]);
			num += m_groups[i]*3;
		}
	}
}

//------------------------------------------------------------------------
//fƂ̓蔻
//posuC̈ʒuv
//diruČv
//distanceuC̈ʒu瓖ʒu܂ł̋v
//nomaluʂ̖@v
//------------------------------------------------------------------------
bool CModel::Collision(D3DXVECTOR3 pos,D3DXVECTOR3 dir,float* dist,D3DXVECTOR3* normal)
{
    D3DXMATRIX mat,matInv;
	D3DXVECTOR4 v;
	multimap < float, int > distance;
	float d;

	//f̃[hs̋ts
    D3DXMatrixInverse( &matInv, NULL, &m_transform );
	//tsC̍WɃgXtH[
	D3DXVec3Transform(&v,&pos,&matInv);
	pos = D3DXVECTOR3(v);

	//f̃[hs񂩂畽sړs̋ts
	mat = m_transform;
	mat._41 = 0;
	mat._42 = 0;
	mat._43 = 0;
    D3DXMatrixInverse( &matInv, NULL, &mat );
	//tsC̕ɃgXtH[
	D3DXVec3Transform(&v,&dir,&matInv);
	dir = D3DXVECTOR3(v);

	//ʂʂgXtH[ĂɁACtsŃgXtH[
	for( int f = 0; f < GetFacesSize(); ++f )
	{
		D3DXVECTOR3 p0,p1,p2;
		CFace* face = &m_faces[f];
		switch ( face->GetIndexSize() )
		{
		case 4:
			p0 = (D3DXVECTOR3)m_vertices[face->GetIndex(0)];
			p1 = (D3DXVECTOR3)m_vertices[face->GetIndex(2)];
			p2 = (D3DXVECTOR3)m_vertices[face->GetIndex(3)];
			//ʂƃCƂ̓蔻
			if ( D3DXIntersectTri(&p0, &p1, &p2, &pos, &dir, NULL, NULL, &d) )
			{
				//ƖʃCfbNXԍ}
				distance.insert(make_pair(d,f));
				break;
			}
		case 3:
			p0 = (D3DXVECTOR3)m_vertices[face->GetIndex(0)];
			p1 = (D3DXVECTOR3)m_vertices[face->GetIndex(1)];
			p2 = (D3DXVECTOR3)m_vertices[face->GetIndex(2)];
			//ʂƃCƂ̓蔻
			if ( D3DXIntersectTri(&p0, &p1, &p2, &pos, &dir, NULL, NULL, &d) )
			{
				//ƖʃCfbNXԍ}
				distance.insert(make_pair(d,f));
			}
			break;
		}
	}

	//qbgʂꍇ
	if ( (int)distance.size() > 0 )
	{
		//}EX̃Cԋ߂ʂƂ̋
		if ( dist != NULL )
		{
			*dist = distance.begin()->first;
		}
		//}EX̃Cԋ߂ʂ̖@
		if ( normal != NULL )
		{
			D3DXVECTOR3 v = m_faces[distance.begin()->second].GetNormal();
			normal->x = -v.x;
			normal->y = -v.y;
			normal->z = -v.z;
		}
		return true;
	}

	return false;
}

//------------------------------------------------------------------------
//fƃ}EX̓蔻(ʔԍԂB-1Ȃ瓖ĂȂ)
//------------------------------------------------------------------------
int CModel::MousePick(int x,int y,LPDIRECT3DDEVICE9 dev,D3DXVECTOR3* pos)
{
    D3DXVECTOR3 vecRayPos;
    D3DXVECTOR3 vecRayDir;
	D3DVIEWPORT9  viewPort;
	D3DXMATRIX matProj,matView;
	multimap < float, int > distance;
	float d;

	dev->GetViewport(&viewPort);
	dev->GetTransform(D3DTS_PROJECTION,&matProj);
	dev->GetTransform(D3DTS_VIEW,&matView);

    D3DXVECTOR3 vDir;
    D3DXVECTOR3 vPickRayDir;
    D3DXVECTOR3 vPickRayOrig;

	//XN[ԂŃsbNC̃xNgvZ
    D3DXVECTOR3 v;
	v.x =  ( ( ( 2.0f * x ) / viewPort.Width  ) - 1 ) / matProj._11;
    v.y = -( ( ( 2.0f * y ) / viewPort.Height ) - 1 ) / matProj._22;
    v.z = -1.0f;

	//[hs̋ts擾
    D3DXMATRIXA16 m1, m;
    m1 = m_transform * matView;
    D3DXMatrixInverse( &m, NULL, &m1 );

	//XN[Ԃ̃sbNC3DԂɃgXtH[
	vDir.x  = v.x*m._11 + v.y*m._21 + v.z*m._31;
    vDir.y  = v.x*m._12 + v.y*m._22 + v.z*m._32;
    vDir.z  = v.x*m._13 + v.y*m._23 + v.z*m._33;
	D3DXVec3Normalize(&vPickRayDir,&vDir);
    vPickRayOrig.x = m._41;
    vPickRayOrig.y = m._42;
    vPickRayOrig.z = m._43;
    
	//ʂƂɃsbNC̍Wƌœ蔻
	for( int f = 0; f < GetFacesSize(); ++f )
	{
		D3DXVECTOR3 p0,p1,p2;
		CFace* face = &m_faces[f];
		switch ( face->GetIndexSize() )
		{
		case 4:
			p0 = m_vertices[face->GetIndex(0)];
			p1 = m_vertices[face->GetIndex(2)];
			p2 = m_vertices[face->GetIndex(3)];
			//ʂƃCƂ̓蔻
			if ( D3DXIntersectTri(&p0, &p1, &p2, &vPickRayOrig, &vPickRayDir, NULL, NULL, &d) )
			{
				//ƖʃCfbNXԍ}
				distance.insert(make_pair(d,f));
				break;
			}
		case 3:
			p0 = m_vertices[face->GetIndex(0)];
			p1 = m_vertices[face->GetIndex(1)];
			p2 = m_vertices[face->GetIndex(2)];
			//ʂƃCƂ̓蔻
			if ( D3DXIntersectTri(&p0, &p1, &p2, &vPickRayOrig, &vPickRayDir, NULL, NULL, &d) )
			{
				//ƖʃCfbNXԍ}
				distance.insert(make_pair(d,f));
			}
			break;
		}
	}

	//qbgʂꍇ
	if ( (int)distance.size() > 0 )
	{
		//}EX̃Cԋ߂ʂƂ̋
		if ( pos != NULL )
		{
			vPickRayOrig += vPickRayDir*distance.begin()->first;
			pos->x = vPickRayOrig.x;
			pos->y = vPickRayOrig.y;
			pos->z = vPickRayOrig.z;
		}
		//}EX̃Cԋ߂ʂ̃CfbNXԍԂ
		return distance.begin()->second;
	}

	return -1;
}

//------------------------------------------------------------------------
//ʃCfbNX̎擾
//------------------------------------------------------------------------
int CModel::GetFaceIndex(int i,int j)
{
	return m_faces[i].GetIndex(j);
}

//------------------------------------------------------------------------
//_W̎擾
//------------------------------------------------------------------------
D3DXVECTOR3 CModel::GetVertexPos(int i)
{
	return m_vertices[i];
}

//------------------------------------------------------------------------
//f̃Zbg
//------------------------------------------------------------------------
void CModel::SetName(std::string name)
{
	m_name = name;
}

//------------------------------------------------------------------------
//f擾
//------------------------------------------------------------------------
std::string CModel::GetName()
{
	return m_name;
}

//------------------------------------------------------------------------
//ʃZbg
//------------------------------------------------------------------------
void CModel::SetMirrorAxis(int mirror_axis)
{
	m_mirror_axis = mirror_axis;
}

//------------------------------------------------------------------------
//ʎ擾
//------------------------------------------------------------------------
int CModel::GetMirrorAxis()
{
	return m_mirror_axis;
}

//------------------------------------------------------------------------
//X[WOpxZbg
//------------------------------------------------------------------------
void CModel::SetFacet(float facet)
{
	m_facet = facet;
}

//------------------------------------------------------------------------
//X[WOpx擾
//------------------------------------------------------------------------
float CModel::GetFacet()
{
	return m_facet;
}

//------------------------------------------------------------------------
//_zɒǉ
//------------------------------------------------------------------------
void CModel::PushBackVertex(const D3DXVECTOR3& v)
{
	m_vertices.push_back(v);
}

//------------------------------------------------------------------------
//ʔzɒǉ
//------------------------------------------------------------------------
void CModel::PushBackFace(const CFace& f)
{
	m_faces.push_back(f);
}

//------------------------------------------------------------------------
//_z̃TCY擾
//------------------------------------------------------------------------
int CModel::GetVerticesSize()
{
	return (int)m_vertices.size();
}

//------------------------------------------------------------------------
//ʔz̃TCY擾
//------------------------------------------------------------------------
int CModel::GetFacesSize()
{
	return (int)m_faces.size();
}

//------------------------------------------------------------------------
//Aj[Vzɒǉ
//------------------------------------------------------------------------
void CModel::PushBackAnimation()
{
	m_animations.resize(m_animations.size()+1);
}

//------------------------------------------------------------------------
//Aj[Vzɒǉ2
//------------------------------------------------------------------------
void CModel::PushBackAnimation(int i,const D3DXMATRIX& m)
{
	m_animations[i].push_back(m);
}

//------------------------------------------------------------------------
//Aj[Vz̃TCY擾
//------------------------------------------------------------------------
int CModel::GetAnimationSize()
{
	return (int)m_animations.size();
}

//------------------------------------------------------------------------
//_obt@̉
//------------------------------------------------------------------------
void CModel::ReleaseVB()
{
	SAFE_RELEASE(m_vb);
}

//------------------------------------------------------------------------
//Zq̃I[o[[h
//------------------------------------------------------------------------
CModel::CModel(const CModel& pt)
{
	*this = pt;
}
